(*
 * F-ing Modules with Coq
 *
 * (c) 2009 Andreas Rossberg & Claudio Russo
 *)

Require Import Metatheory.
Require Import AdditionalTactics.
Require Import Misc.
Require Import Fomega.
Require Import Derived_multiple.


(*************************************************************************)
(* Tuples *)
(*************************************************************************)

(* t1 * ... * tN   :=   t1 * (... * (tN * 1)...) *)
Fixpoint tupT ts {struct ts} : typ :=
  match ts with
  | nil => unitT
  | t::ts' => timT t (tupT ts')
  end.

(* (e1,...,eN)   :=   (e1,(...,(eN,())...) *)
Fixpoint tupE es {struct es} : exp :=
  match es with
  | nil => trivE
  | e::es' => pairE e (tupE es')
  end.

(* e.n   :=   e.2^(n-1).1 *)
Fixpoint projE' n e {struct n} : exp :=
  match n with
  | O => fstE e
  | S n' => projE' n' (sndE e)
  end.
Definition projE n e : exp :=
  match n with
  | O => trivE
  | S n' => projE' n' e
  end.


(*************************************************************************)
(* Substitution *)
(*************************************************************************)

Lemma substAT_tup : forall a t' ts,
  substAT a t' (tupT ts) = tupT (map (substAT a t') ts).
Proof.
  intros a t' ts.
  induction ts; simpl tupT; simpl substAT.
    Case "ts = nil".
      reflexivity.
    Case "ts = t::ts'".
      rewrite IHts. reflexivity.
Qed.

Corollary openAT_tup : forall t' ts,
  openAT t' (tupT ts) = tupT (map (openAT t') ts).
Proof.
  intros t' ts. unfold openAT. rewrite substAT_tup. unfold substAT. reflexivity.
Qed.

Lemma substAE_tup : forall a t es,
  substAE a t (tupE es) = tupE (map (substAE a t) es).
Proof.
  intros a t es.
  induction es; simpl tupE; simpl substAE.
    Case "es = nil".
      reflexivity.
    Case "es = e::es'".
      rewrite IHes. reflexivity.
Qed.

Lemma openAE_tup : forall t es,
  openAE t (tupE es) = tupE (map (openAE t) es).
Proof.
  intros t es. unfold openAE. rewrite substAE_tup. unfold substAE. reflexivity.
Qed.

Lemma substXE_tup : forall x e' es,
  substXE x e' (tupE es) = tupE (map (substXE x e') es).
Proof.
  intros x e' es.
  induction es; simpl tupE; simpl substXE.
    Case "es = nil".
      reflexivity.
    Case "es = e::es'".
      rewrite IHes. reflexivity.
Qed.

Lemma substAE_proj' : forall a t n e,
  substAE a t (projE' n e) = projE' n (substAE a t e).
Proof.
  intros a t n e.
  generalize dependent e.
  induction n; intros e; simpl projE'; simpl substAE.
    Case "n = 0".
      reflexivity.
    Case "n = S n'".
      rewrite IHn. reflexivity.
Qed.

Corollary substAE_proj : forall a t n e,
  substAE a t (projE n e) = projE n (substAE a t e).
Proof.
  intros a t n e.
  destruct n; unfold projE; simpl substAE; auto using substAE_proj'.
Qed.

Corollary openAE_proj : forall t n e,
  openAE t (projE n e) = projE n (openAE t e).
Proof.
  intros t n e. unfold openAE. apply substAE_proj.
Qed.

Lemma substXE_proj' : forall x e' n e,
  substXE x e' (projE' n e) = projE' n (substXE x e' e).
Proof.
  intros x e' n e.
  generalize dependent e.
  induction n; intros e; simpl projE'; simpl substXE.
    Case "n = 0".
      reflexivity.
    Case "n = S n'".
      rewrite IHn. reflexivity.
Qed.

Corollary substXE_proj : forall x e' n e,
  substXE x e' (projE n e) = projE n (substXE x e' e).
Proof.
  intros x e' n e.
  destruct n; unfold projE; simpl substXE; auto using substXE_proj'.
Qed.

Corollary openXE_proj : forall e' n e,
  openXE e' (projE n e) = projE n (openXE e' e).
Proof.
  intros e' n e. unfold openXE. apply substXE_proj.
Qed.


(*************************************************************************)
(* Local closure *)
(*************************************************************************)

Lemma lclosedT_tup : forall ts,
  lclosedTs ts ->
  lclosedT (tupT ts).
Proof.
  induction ts; intro Hts; simpl tupT.
    apply lclosedT_unit.
    apply lclosedT_tim.
      eapply lclosedTs_head. apply Hts.
      apply IHts. eapply lclosedTs_tail. apply Hts.
Qed.

Lemma lclosedE_tup : forall es,
  lclosedEs es ->
  lclosedE (tupE es).
Proof.
  induction es; intro Hes; simpl tupE.
    apply lclosedE_triv.
    apply lclosedE_pair.
      eapply lclosedEs_head. apply Hes.
      apply IHes. eapply lclosedEs_tail. apply Hes.
Qed.

Lemma lclosedE_proj' : forall n e,
  lclosedE e->
  lclosedE (projE' n e).
Proof.
  induction n; intros e He; simpl projE'.
    apply lclosedE_fst. apply He.
    apply IHn. apply lclosedE_snd. apply He.
Qed.

Corollary lclosedE_proj : forall n e,
  lclosedE e ->
  lclosedE (projE n e).
Proof.
  intros n e He. unfold projE.
  destruct n.
    apply lclosedE_triv.
    apply lclosedE_proj'. apply He.
Qed.


(*************************************************************************)
(* Typing *)
(*************************************************************************)

Lemma ofT_tup : forall D ts,
  okD D ->
  (forall t, In t ts -> ofT D t omK) ->
  ofT D (tupT ts) omK.
Proof.
  intros D ts HD H.
  induction ts; simpl tupT in *.
    Case "ts = nil".
      apply ofT_unit. apply HD.
    Case "ts = t :: ts'".
      apply ofT_tim.
        apply H. left. reflexivity.
        apply IHts. intros t Ht. apply H. right. apply Ht.
Qed.

Lemma eqT_tup : forall ts ts',
  length ts = length ts' ->
  (forall t t', In (t, t') (combine ts ts') -> eqT t t') ->
  eqT (tupT ts) (tupT ts').
Proof.
  induction ts; intros ts' Len Hts.
    Case "ts = nil".
      destruct ts'; simpl tupT.
        SCase "ts' = nil".
          apply eqT_unit.
        SCase "ts' <> nil".
          simpl length in Len. inversion Len.
    Case "ts = t::ts".
      destruct ts'; simpl tupT.
        SCase "ts' = nil".
          simpl length in Len. inversion Len.
        SCase "ts' = t'::ts'".
          apply eqT_tim.
            apply Hts. simpl combine. apply in_eq.
            apply IHts.
              simpl length in Len. inversion Len. reflexivity.
              intros ti ti' Elem. apply Hts. simpl combine. apply in_cons. apply Elem.
Qed.

Lemma ofE_tup : forall D G es ts,
  okG D G ->
  length es = length ts ->
  (forall e t, In (e,t) (combine es ts) -> ofE D G e t) ->
  ofE D G (tupE es) (tupT ts).
Proof.
  intros D G es ts HG Hts He. unfold tupE.
  generalize dependent ts.
  induction es; intros ts Hts He; simpl in *.
    Case "es = nil".
      unfold List.length in Hts. destruct ts.
        SCase "ts = nil".
          unfold tupT; simpl. apply ofE_triv. apply HG.
        SCase "ts <> nil".
          inversion Hts.
    Case "es = e::es'".
      destruct ts; simpl in *.
        SCase "ts = nil".
          inversion Hts.
        SCase "ts = t::ts'".
          apply ofE_pair.
            apply He. left. reflexivity.
            apply IHes.
              inversion Hts. reflexivity.
              intros e t0 Het0. apply He. right. apply Het0.
Qed.

Lemma ofE_proj : forall D G e ts n,
  0 <= n < length ts ->
  ofE D G e (tupT ts) ->
  ofE D G (projE (S n) e) (nth n ts unitT).
Proof.
  intros D G e ts n Hn He. unfold projE, nth.
  generalize dependent ts. generalize dependent e.
  induction n; intros e ts Hn He; destruct ts; simpl in *.
    Case "n = 0, ts = nil".
      inversion Hn. inversion H0.
    Case "n = 0, ts = t::ts'".
      eapply ofE_fst. apply He.
    Case "n = S n', ts = nil".
      inversion Hn. inversion H0.
    Case "n = S n', ts = t::ts'".
      apply IHn.
        split. auto with arith. inversion Hn. apply Arith.Lt.lt_S_n. apply H0.
        eapply ofE_snd. apply He.
Qed.
