(*
 * F-ing Modules with Coq
 *
 * (c) 2009 Andreas Rossberg & Claudio Russo
 *)

Require Import Metatheory.
Require Import Bool_nat.
Require Import Fomega_syntax.


(*************************************************************************)
(* Free variables *)
(*************************************************************************)

Fixpoint varsAT (t : typ) {struct t} : vars :=
  match t with
  | varT (free a) => {{a}}
  | varT (bnd _) => {}
  | unitT => {}
  | arrT t1 t2 => (varsAT t1) \u (varsAT t2)
  | timT t1 t2 => (varsAT t1) \u (varsAT t2)
  | plusT t1 t2 => (varsAT t1) \u (varsAT t2)
  | recT k1 t2 => varsAT t2
  | allT k1 t2 => varsAT t2
  | exT  k1 t2 => varsAT t2
  | trivT => {}
  | lamT k1 t2 => varsAT t2
  | appT t1 t2 => (varsAT t1) \u (varsAT t2)
  | pairT t1 t2 => (varsAT t1) \u (varsAT t2)
  | fstT t1 => varsAT t1
  | sndT t1 => varsAT t1
  end.

Fixpoint varsAE (e : exp) {struct e} : vars :=
  match e with
  | varE x => {}
  | trivE => {}
  | lamE t1 e2 => (varsAT t1) \u (varsAE e2)
  | appE e1 e2 => (varsAE e1) \u (varsAE e2)
  | pairE e1 e2 => (varsAE e1) \u (varsAE e2)
  | fstE e1 => varsAE e1
  | sndE e1 => varsAE e1
  | lftE e1 t2 => (varsAE e1) \u (varsAT t2)
  | rgtE e1 t2 => (varsAE e1) \u (varsAT t2)
  | caseE e1 e2 e3 => (varsAE e1) \u (varsAE e2) \u (varsAE e3)
  | rollE e1 t2 => (varsAE e1) \u (varsAT t2)
  | unrollE e1 => varsAE e1
  | genE k1 e2 => varsAE e2
  | instE e1 t2 => (varsAE e1) \u (varsAT t2)
  | packE t1 e2 t3 => (varsAT t1) \u (varsAE e2) \u (varsAT t3)
  | unpackE e1 e2 => (varsAE e1) \u (varsAE e2)
  end.

Fixpoint varsXE (e : exp) {struct e} : vars :=
  match e with
  | varE (free x) => {{x}}
  | varE (bnd _) => {}
  | trivE => {}
  | lamE t1 e2 => varsXE e2
  | appE e1 e2 => (varsXE e1) \u (varsXE e2)
  | pairE e1 e2 => (varsXE e1) \u (varsXE e2)
  | fstE e1 => varsXE e1
  | sndE e1 => varsXE e1
  | lftE e1 t2 => varsXE e1
  | rgtE e1 t2 => varsXE e1
  | caseE e1 e2 e3 => (varsXE e1) \u (varsXE e2) \u (varsXE e3)
  | rollE e1 t2 => varsXE e1
  | unrollE e1 => varsXE e1
  | genE k1 e2 => varsXE e2
  | instE e1 t2 => varsXE e1
  | packE t1 e2 t3 => varsXE e2
  | unpackE e1 e2 => (varsXE e1) \u (varsXE e2)
  end.

Definition varsAG (G : eenv) :=
  fold_right S.union S.empty (map (fun xt => varsAT (snd xt)) G).

Definition varsT t := varsAT t.
Definition varsE e := varsAE e \u varsXE e.
Definition varsG G := varsAG G.


(*************************************************************************)
(* Substitution *)
(*************************************************************************)

Definition free_inject (x y : var) (H : free x = free y) : x = y.
Proof.
  intros. injection H. trivial.
Defined.

Definition bnd_inject (i j : nat) (H : bnd i = bnd j) : i = j.
Proof.
  intros. injection H. trivial.
Defined.

(*
Definition free_bnd (a : var) (i : nat) : free a <> bnd i.
Proof.
  unfold not. intros. inversion H.
Defined.
*)
Definition free_bnd (a : var) (i : nat) : free a <> bnd i :=
  fun H : free a = bnd i =>
  match H in (_ = x) return (match x with free _ => True | bnd _ => False end) with
  | refl_equal => I
  end.


Definition eqOcc (x : occ) (y : occ) : {x = y} + {x <> y} :=
  match x, y return {x = y} + {x <> y} with
  | free a, free b =>
    match a == b with
    | left P => left (free a <> free b) (f_equal free P)
    | right P => right (free a = free b) (fun H : free a = free b => P (free_inject a b H))
    end
  | bnd i, bnd j =>
    match i === j with
    | left P => left (bnd i <> bnd j) (f_equal bnd P)
    | right P => right (bnd i = bnd j) (fun H : bnd i = bnd j => P (bnd_inject i j H))
    end
  | free a, bnd i => right (free a = bnd i) (free_bnd a i)
  | bnd i, free a => right (bnd i = free a) (fun H => free_bnd a i (sym_eq H))
  end.

Definition shift (x : occ) : occ :=
  match x with
  | free a => free a
  | bnd i => bnd (S i)
  end.

Fixpoint substAT (a' : occ) (t' : typ) (t : typ) {struct t} : typ :=
  match t with
  | varT a => if eqOcc a a' then t' else varT a
  | unitT => unitT
  | arrT t1 t2 => arrT (substAT a' t' t1) (substAT a' t' t2)
  | timT t1 t2 => timT (substAT a' t' t1) (substAT a' t' t2)
  | plusT t1 t2 => plusT (substAT a' t' t1) (substAT a' t' t2)
  | recT k1 t2 => recT k1 (substAT (shift a') t' t2)
  | allT k1 t2 => allT k1 (substAT (shift a') t' t2)
  | exT  k1 t2 => exT k1 (substAT (shift a') t' t2)
  | trivT => trivT
  | lamT k1 t2 => lamT k1 (substAT (shift a') t' t2)
  | appT t1 t2 => appT (substAT a' t' t1) (substAT a' t' t2)
  | pairT t1 t2 => pairT (substAT a' t' t1) (substAT a' t' t2)
  | fstT t1 => fstT (substAT a' t' t1)
  | sndT t1 => sndT (substAT a' t' t1)
  end.

Fixpoint substAE (a' : occ) (t' : typ) (e : exp) {struct e} : exp :=
  match e with
  | varE x => varE x
  | trivE => trivE
  | lamE t1 e2 => lamE (substAT a' t' t1) (substAE a' t' e2)
  | appE e1 e2 => appE (substAE a' t' e1) (substAE a' t' e2)
  | pairE e1 e2 => pairE (substAE a' t' e1) (substAE a' t' e2)
  | fstE e1 => fstE (substAE a' t' e1)
  | sndE e1 => sndE (substAE a' t' e1)
  | lftE e1 t2 => lftE (substAE a' t' e1) (substAT a' t' t2)
  | rgtE e1 t2 => rgtE (substAE a' t' e1) (substAT a' t' t2)
  | caseE e1 e2 e3 => caseE (substAE a' t' e1) (substAE a' t' e2) (substAE a' t' e3)
  | rollE e1 t2 => rollE (substAE a' t' e1) (substAT a' t' t2)
  | unrollE e1 => unrollE (substAE a' t' e1)
  | genE k1 e2 => genE k1 (substAE (shift a') t' e2)
  | instE e1 t2 => instE (substAE a' t' e1) (substAT a' t' t2)
  | packE t1 e2 t3 => packE (substAT a' t' t1) (substAE a' t' e2) (substAT a' t' t3)
  | unpackE e1 e2 => unpackE (substAE a' t' e1) (substAE (shift a') t' e2)
  end.

Fixpoint substXE (x' : occ) (e' : exp) (e : exp) {struct e} : exp :=
  match e with
  | varE x => if eqOcc x x' then e' else varE x
  | trivE => trivE
  | lamE t1 e2 => lamE t1 (substXE (shift x') e' e2)
  | appE e1 e2 => appE (substXE x' e' e1) (substXE x' e' e2)
  | pairE e1 e2 => pairE (substXE x' e' e1) (substXE x' e' e2)
  | fstE e1 => fstE (substXE x' e' e1)
  | sndE e1 => sndE (substXE x' e' e1)
  | lftE e1 t2 => lftE (substXE x' e' e1) t2
  | rgtE e1 t2 => rgtE (substXE x' e' e1) t2
  | caseE e1 e2 e3 => caseE (substXE x' e' e1) (substXE (shift x') e' e2) (substXE (shift x') e' e3)
  | rollE e1 t2 => rollE (substXE x' e' e1) t2
  | unrollE e1 => unrollE (substXE x' e' e1)
  | genE k1 e2 => genE k1 (substXE x' e' e2)
  | instE e1 t2 => instE (substXE x' e' e1) t2
  | packE t1 e2 t3 => packE t1 (substXE x' e' e2) t3
  | unpackE e1 e2 => unpackE (substXE x' e' e1) (substXE (shift x') e' e2)
  end.

Definition substAG (a : occ) (t' : typ) := Env.map (substAT a t').

Definition openAT := substAT 0.
Definition openAE := substAE 0.
Definition openXE := substXE 0.


(*************************************************************************)
(* Local closure *)
(*************************************************************************)

Inductive lclosedT : typ -> Prop :=
  | lclosedT_var : forall a,
      lclosedT (varT (free a))
  | lclosedT_unit :
      lclosedT (unitT)
  | lclosedT_arr : forall t1 t2,
      lclosedT t1 ->
      lclosedT t2 ->
      lclosedT (arrT t1 t2)
  | lclosedT_tim : forall t1 t2,
      lclosedT t1 ->
      lclosedT t2 ->
      lclosedT (timT t1 t2)
  | lclosedT_plus : forall t1 t2,
      lclosedT t1 ->
      lclosedT t2 ->
      lclosedT (plusT t1 t2)
  | lclosedT_rec : forall L k1 t2,
      (forall a : var, a \notin L -> lclosedT (openAT a t2)) ->
      lclosedT (recT k1 t2)
  | lclosedT_all : forall L k1 t2,
      (forall a : var, a \notin L -> lclosedT (openAT a t2)) ->
      lclosedT (allT k1 t2)
  | lclosedT_ex : forall L k1 t2,
      (forall a : var, a \notin L -> lclosedT (openAT a t2)) ->
      lclosedT (exT k1 t2)
  | lclosedT_triv :
      lclosedT (trivT)
  | lclosedT_lam : forall L k1 t2,
      (forall a : var, a \notin L -> lclosedT (openAT a t2)) ->
      lclosedT (lamT k1 t2)
  | lclosedT_app : forall t1 t2,
      lclosedT t1 ->
      lclosedT t2 ->
      lclosedT (appT t1 t2)
  | lclosedT_pair : forall t1 t2,
      lclosedT t1 ->
      lclosedT t2 ->
      lclosedT (pairT t1 t2)
  | lclosedT_fst : forall t1,
      lclosedT t1 ->
      lclosedT (fstT t1)
  | lclosedT_snd : forall t1,
      lclosedT t1 ->
      lclosedT (sndT t1)
.

Inductive lclosedE : exp -> Prop :=
  | lclosedE_var : forall x,
      lclosedE (varE (free x))
  | lclosedE_triv :
      lclosedE (trivE)
  | lclosedE_lam : forall t1 e2 L,
      lclosedT t1 ->
      (forall x : var, x \notin L -> lclosedE (openXE x e2)) ->
      lclosedE (lamE t1 e2)
  | lclosedE_app : forall e1 e2,
      lclosedE e1 ->
      lclosedE e2 ->
      lclosedE (appE e1 e2)
  | lclosedE_pair : forall e1 e2,
      lclosedE e1 ->
      lclosedE e2 ->
      lclosedE (pairE e1 e2)
  | lclosedE_fst : forall e1,
      lclosedE e1 ->
      lclosedE (fstE e1)
  | lclosedE_snd : forall e1,
      lclosedE e1 ->
      lclosedE (sndE e1)
  | lclosedE_lft : forall e1 t2,
      lclosedE e1 ->
      lclosedT t2 ->
      lclosedE (lftE e1 t2)
  | lclosedE_rgt : forall e1 t2,
      lclosedE e1 ->
      lclosedT t2 ->
      lclosedE (rgtE e1 t2)
  | lclosedE_case : forall e1 e2 e3 L,
      lclosedE e1 ->
      (forall x : var, x \notin L -> lclosedE (openXE x e2)) ->
      (forall x : var, x \notin L -> lclosedE (openXE x e3)) ->
      lclosedE (caseE e1 e2 e3)
  | lclosedE_roll : forall e1 t2,
      lclosedE e1 ->
      lclosedT t2 ->
      lclosedE (rollE e1 t2)
  | lclosedE_unroll : forall e1,
      lclosedE e1 ->
      lclosedE (unrollE e1)
  | lclosedE_gen : forall k1 e2 L,
      (forall a : var, a \notin L -> lclosedE (openAE a e2)) ->
      lclosedE (genE k1 e2)
  | lclosedE_inst : forall e1 t2,
      lclosedE e1 ->
      lclosedT t2 ->
      lclosedE (instE e1 t2)
  | lclosedE_pack : forall t1 e2 t3,
      lclosedT t1 ->
      lclosedE e2 ->
      lclosedT t3 ->
      lclosedE (packE t1 e2 t3)
  | lclosedE_unpack : forall e1 e2 L,
      lclosedE e1 ->
      (forall a x : var, a \notin L -> x \notin L -> lclosedE (openXE x (openAE a e2))) ->
      lclosedE (unpackE e1 e2)
.

Hint Constructors lclosedT.
Hint Constructors lclosedE.
